"""pdsx - general-purpose cli for atproto record operations."""

from __future__ import annotations

import importlib.metadata

try:
    __version__ = importlib.metadata.version("pdsx")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"  # fallback for development mode

__all__ = ["__version__"]



"""configuration management for pdsx."""

from __future__ import annotations

import warnings
from pathlib import Path

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

# suppress pydantic warning about Field defaults
warnings.filterwarnings(
    "ignore", category=UserWarning, module="pydantic._internal._generate_schema"
)


class Settings(BaseSettings):
    """settings for atproto cli."""

    model_config = SettingsConfigDict(
        env_file=str(Path.cwd() / ".env"),
        env_file_encoding="utf-8",
        extra="ignore",
        case_sensitive=False,
    )

    atproto_pds_url: str = Field(
        default="https://bsky.social",
        description="PDS URL (only used for unauthenticated reads with -r flag)",
    )
    atproto_handle: str = Field(default="", description="Your atproto handle")
    atproto_password: str = Field(default="", description="Your atproto app password")


settings = Settings()



"""tests for configuration."""

from __future__ import annotations

import os

from pdsx._internal.config import Settings


def test_settings_defaults() -> None:
    """test default settings values when no env vars are set."""
    # clear env vars for this test
    old_values = {}
    for key in ["ATPROTO_PDS_URL", "ATPROTO_HANDLE", "ATPROTO_PASSWORD"]:
        old_values[key] = os.environ.pop(key, None)

    try:
        settings = Settings(_env_file=None)
        assert settings.atproto_pds_url == "https://bsky.social"
        assert settings.atproto_handle == ""
        assert settings.atproto_password == ""
    finally:
        # restore env vars
        for key, value in old_values.items():
            if value is not None:
                os.environ[key] = value


def test_settings_override() -> None:
    """test overriding settings."""
    settings = Settings(
        atproto_pds_url="https://custom.pds",
        atproto_handle="test.handle",
        atproto_password="test_password",
    )
    assert settings.atproto_pds_url == "https://custom.pds"
    assert settings.atproto_handle == "test.handle"
    assert settings.atproto_password == "test_password"



"""atproto record operations."""

from __future__ import annotations

import sys
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING

if sys.version_info >= (3, 11):
    from datetime import UTC
else:
    UTC = timezone.utc

if TYPE_CHECKING:
    from atproto import AsyncClient, models

from pdsx._internal.resolution import URIParts
from pdsx._internal.types import RecordValue


async def list_records(
    client: AsyncClient,
    collection: str,
    limit: int,
    repo: str | None = None,
    cursor: str | None = None,
) -> models.ComAtprotoRepoListRecords.Response:
    """list records in a collection.

    Args:
        client: authenticated atproto client
        collection: collection name
        limit: maximum number of records
        repo: optional repo DID (defaults to authenticated user)
        cursor: optional pagination cursor

    Returns:
        response with records and optional cursor for next page
    """
    target_repo = repo or (client.me.did if client.me else None)
    if not target_repo:
        raise ValueError("no repo specified and not authenticated")

    params: dict[str, str | int] = {
        "repo": target_repo,
        "collection": collection,
        "limit": limit,
    }

    if cursor:
        params["cursor"] = cursor

    return await client.com.atproto.repo.list_records(params)


async def get_record(
    client: AsyncClient,
    uri: str,
) -> models.ComAtprotoRepoGetRecord.Response:
    """get a specific record.

    Args:
        client: authenticated atproto client
        uri: record AT-URI (can be shorthand like 'collection/rkey' if authenticated)

    Returns:
        record response
    """
    parts = URIParts.from_uri(uri, client.me.did if client.me else None)

    return await client.com.atproto.repo.get_record(
        {
            "repo": parts.repo,
            "collection": parts.collection,
            "rkey": parts.rkey,
        }
    )


async def create_record(
    client: AsyncClient,
    collection: str,
    record: dict[str, RecordValue],
) -> models.ComAtprotoRepoCreateRecord.Response:
    """create a new record.

    Args:
        client: authenticated atproto client
        collection: collection name
        record: record data

    Returns:
        created record response
    """
    if client.me is None:
        raise ValueError("not authenticated")

    # ensure $type is set
    if "$type" not in record:
        record["$type"] = collection

    # add createdAt if not present
    if "createdAt" not in record:
        record["createdAt"] = datetime.now(UTC).isoformat().replace("+00:00", "Z")

    return await client.com.atproto.repo.create_record(
        {
            "repo": client.me.did,
            "collection": collection,
            "record": record,
        }
    )


async def update_record(
    client: AsyncClient,
    uri: str,
    updates: dict[str, RecordValue],
) -> models.ComAtprotoRepoPutRecord.Response:
    """update an existing record.

    Args:
        client: authenticated atproto client
        uri: record AT-URI (can be shorthand like 'collection/rkey' if authenticated)
        updates: fields to update

    Returns:
        updated record response
    """
    parts = URIParts.from_uri(uri, client.me.did if client.me else None)

    # get current
    current = await client.com.atproto.repo.get_record(
        {
            "repo": parts.repo,
            "collection": parts.collection,
            "rkey": parts.rkey,
        }
    )

    # merge
    if isinstance(current.value, dict):
        updated_value = {**current.value, **updates}
    else:
        updated_value = dict(current.value)
        updated_value.update(updates)

    # put
    return await client.com.atproto.repo.put_record(
        {
            "repo": parts.repo,
            "collection": parts.collection,
            "rkey": parts.rkey,
            "record": updated_value,
        }
    )


async def delete_record(
    client: AsyncClient,
    uri: str,
) -> None:
    """delete a record.

    Args:
        client: authenticated atproto client
        uri: record AT-URI (can be shorthand like 'collection/rkey' if authenticated)
    """
    parts = URIParts.from_uri(uri, client.me.did if client.me else None)

    await client.com.atproto.repo.delete_record(
        {
            "repo": parts.repo,
            "collection": parts.collection,
            "rkey": parts.rkey,
        }
    )


async def upload_blob(
    client: AsyncClient,
    file_path: str | Path,
) -> models.ComAtprotoRepoUploadBlob.Response:
    """upload a blob (image, video, etc.).

    Args:
        client: authenticated atproto client
        file_path: path to file to upload

    Returns:
        upload response with blob reference
    """
    if client.me is None:
        raise ValueError("not authenticated")

    file_path = Path(file_path)
    if not file_path.exists():
        raise FileNotFoundError(f"file not found: {file_path}")

    # read file bytes
    data = file_path.read_bytes()

    # upload blob
    return await client.com.atproto.repo.upload_blob(data)



"""general-purpose cli for atproto record operations."""

from __future__ import annotations

import argparse
import asyncio
import sys
import warnings
from typing import NoReturn

# suppress pydantic warnings from atproto library
warnings.filterwarnings("ignore", category=UserWarning, module="pydantic")

from atproto import AsyncClient  # noqa: E402

from pdsx import __version__  # noqa: E402
from pdsx._internal.auth import login  # noqa: E402
from pdsx._internal.config import settings  # noqa: E402
from pdsx._internal.display import (  # noqa: E402
    console,
    display_record,
    display_records,
    display_success,
)
from pdsx._internal.operations import (  # noqa: E402
    create_record,
    delete_record,
    get_record,
    list_records,
    update_record,
    upload_blob,
)
from pdsx._internal.output import OutputFormat  # noqa: E402
from pdsx._internal.parsing import parse_key_value_args  # noqa: E402
from pdsx._internal.resolution import discover_pds  # noqa: E402
from pdsx._internal.types import RecordValue  # noqa: E402


async def cmd_list(
    client: AsyncClient,
    collection: str,
    limit: int,
    repo: str | None = None,
    cursor: str | None = None,
    output_format: OutputFormat | None = None,
) -> None:
    """list records in a collection."""
    response = await list_records(client, collection, limit, repo, cursor)

    # determine output format - default to compact (most readable for most data)
    fmt = output_format or OutputFormat.COMPACT

    display_records(collection, response.records, output_format=fmt)

    # display cursor if there are more pages
    # for structured output formats (json/yaml), send to stderr to avoid breaking parsing
    if response.cursor:
        structured_formats = (OutputFormat.JSON, OutputFormat.YAML)
        if fmt in structured_formats:
            # use stderr for structured formats to avoid breaking json/yaml parsing
            print(f"\nnext page cursor: {response.cursor}", file=sys.stderr)
        else:
            console.print(f"\n[dim]next page cursor:[/dim] {response.cursor}")


async def cmd_get(
    client: AsyncClient, uri: str, output_format: OutputFormat | None = None
) -> None:
    """get a specific record."""
    response = await get_record(client, uri)
    fmt = output_format or OutputFormat.TABLE
    display_record(response, output_format=fmt)


async def cmd_create(
    client: AsyncClient,
    collection: str,
    record: dict[str, RecordValue],
) -> None:
    """create a new record."""
    response = await create_record(client, collection, record)
    display_success("created", response.uri, response.cid, collection)


async def cmd_update(
    client: AsyncClient,
    uri: str,
    updates: dict[str, RecordValue],
) -> None:
    """update an existing record."""
    response = await update_record(client, uri, updates)
    display_success("updated", response.uri, response.cid)


async def cmd_delete(client: AsyncClient, uri: str) -> None:
    """delete a record."""
    await delete_record(client, uri)
    display_success("deleted", "", "")


async def cmd_upload_blob(client: AsyncClient, file_path: str) -> None:
    """upload a blob (image, video, etc.)."""
    response = await upload_blob(client, file_path)

    # display blob reference in json format for easy copying
    import json

    blob_ref = {
        "$type": "blob",
        "ref": {"$link": response.blob.ref.link},
        "mimeType": response.blob.mime_type,
        "size": response.blob.size,
    }
    console.print("[green]✓[/green] blob uploaded successfully")
    console.print("\n[bold]blob reference:[/bold]")
    console.print(json.dumps(blob_ref, indent=2))
    console.print(
        "\n[dim]use this blob reference in records (e.g., for post embeds)[/dim]"
    )


async def async_main() -> int:
    """main entry point."""
    parser = argparse.ArgumentParser(
        description="atproto record operations",
        epilog="""
examples:
  # read anyone's posts (no auth needed)
  pdsx -r zzstoatzz.io ls app.bsky.feed.post

  # read with DID (more durable than handle)
  pdsx -r did:plc:abc123 ls app.bsky.feed.post

  # list your own records (requires auth)
  pdsx --handle you.bsky.social --password xxxx-xxxx ls app.bsky.feed.post

  # create a record (requires auth)
  pdsx --handle you.bsky.social create app.bsky.feed.post text='hello'

note: -r flag goes BEFORE the command (ls, get, etc.)
      auth flags (--handle, --password) also go BEFORE the command
        """,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    # version flag
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"pdsx {__version__}",
    )

    # global identity flag - can be handle or DID
    parser.add_argument(
        "-r",
        "--repo",
        metavar="REPO",
        help="read from another repo (handle or DID) - no auth needed for public data",
    )

    # auth options (only needed for writes to your own repo)
    parser.add_argument(
        "--handle",
        help="your atproto handle for authentication (required for writes)",
    )
    parser.add_argument(
        "--password",
        help="your atproto app password (get from Bluesky settings)",
    )
    parser.add_argument(
        "--pds",
        help="custom PDS URL (auto-discovered from handle if not specified)",
    )

    subparsers = parser.add_subparsers(dest="command", help="command")

    # list (ls alias)
    list_parser = subparsers.add_parser("list", aliases=["ls"], help="list records")
    list_parser.add_argument(
        "collection", help="collection name (e.g., app.bsky.feed.post)"
    )
    list_parser.add_argument("--limit", type=int, default=50, help="max records")
    list_parser.add_argument(
        "--cursor",
        help="pagination cursor from previous response",
    )
    list_parser.add_argument(
        "-o",
        "--output",
        choices=["json", "yaml", "table", "compact"],
        help="output format (default: compact)",
    )

    # get (cat alias)
    get_parser = subparsers.add_parser("get", aliases=["cat"], help="get record")
    get_parser.add_argument("uri", help="record AT-URI")
    get_parser.add_argument(
        "-o",
        "--output",
        choices=["json", "yaml", "table", "compact"],
        help="output format (default: table)",
    )

    # create (touch/add aliases)
    create_parser = subparsers.add_parser(
        "create", aliases=["touch", "add"], help="create record"
    )
    create_parser.add_argument("collection", help="collection name")
    create_parser.add_argument(
        "fields",
        nargs="+",
        help="record fields as key=value pairs (e.g., title='My Song' artist='Artist')",
    )

    # update (edit alias)
    update_parser = subparsers.add_parser(
        "update", aliases=["edit"], help="update record"
    )
    update_parser.add_argument("uri", help="record AT-URI")
    update_parser.add_argument(
        "fields", nargs="+", help="fields to update as key=value pairs"
    )

    # delete (rm alias)
    delete_parser = subparsers.add_parser(
        "delete", aliases=["rm"], help="delete record"
    )
    delete_parser.add_argument("uri", help="record AT-URI")

    # upload-blob
    upload_blob_parser = subparsers.add_parser(
        "upload-blob", help="upload a blob (image, video, etc.)"
    )
    upload_blob_parser.add_argument("file_path", help="path to file to upload")

    args = parser.parse_args()

    if not args.command:
        parser.print_help()
        return 1

    # update pds if provided
    if args.pds:
        settings.atproto_pds_url = args.pds

    try:
        # determine if auth is needed
        # reads with --repo don't need auth, writes always need auth
        read_commands = ("list", "ls", "get", "cat")
        is_read = args.command in read_commands
        has_repo_target = args.repo is not None

        # auth required if: doing a write operation OR doing a read without --repo
        auth_needed = not is_read or not has_repo_target

        # create client with or without base_url depending on auth
        if auth_needed:
            # for authenticated operations, let SDK discover PDS from handle
            # unless user explicitly provided --pds flag
            client = AsyncClient(base_url=args.pds) if args.pds else AsyncClient()

            await login(
                client,
                args.handle or settings.atproto_handle,
                args.password or settings.atproto_password,
                silent=True,
                required=True,
            )
        else:
            # for unauthenticated reads with -r, auto-discover PDS from handle/DID
            pds_url = (
                await discover_pds(args.repo) if args.repo else settings.atproto_pds_url
            )
            client = AsyncClient(base_url=pds_url)

        if args.command in ("list", "ls"):
            output_fmt = OutputFormat(args.output) if args.output else None
            await cmd_list(
                client,
                args.collection,
                args.limit,
                args.repo,
                args.cursor,
                output_format=output_fmt,
            )

        elif args.command in ("get", "cat"):
            output_fmt = (
                OutputFormat[args.output.upper()] if args.output else OutputFormat.TABLE
            )
            await cmd_get(client, args.uri, output_format=output_fmt)

        elif args.command in ("create", "touch", "add"):
            record = parse_key_value_args(args.fields)
            await cmd_create(client, args.collection, record)

        elif args.command in ("update", "edit"):
            updates = parse_key_value_args(args.fields)
            await cmd_update(client, args.uri, updates)

        elif args.command in ("delete", "rm"):
            await cmd_delete(client, args.uri)

        elif args.command == "upload-blob":
            await cmd_upload_blob(client, args.file_path)

        return 0

    except Exception as e:
        console.print(f"[red]error:[/red] {e}")
        import os

        if os.getenv("DEBUG"):
            raise
        return 1


def main() -> NoReturn:
    """synchronous entry point."""
    sys.exit(asyncio.run(async_main()))



"""tests for type aliases."""

from __future__ import annotations

from pdsx._internal.types import RecordValue


def test_record_value_type_alias() -> None:
    """test that RecordValue accepts valid types."""
    valid_values: list[RecordValue] = [
        "string",
        42,
        3.14,
        True,
        False,
        None,
    ]
    assert len(valid_values) == 6



"""argument parsing utilities."""

from __future__ import annotations

import json
import sys
from typing import Any

from rich.console import Console

console = Console()


def parse_key_value_args(args: list[str]) -> dict[str, Any]:
    """parse key=value arguments into a dict.

    supports JSON for complex values (objects/arrays).

    Args:
        args: list of key=value strings

    Returns:
        parsed dictionary

    Examples:
        >>> parse_key_value_args(["name=test", "count=5", "active=true"])
        {'name': 'test', 'count': 5, 'active': True}
        >>> parse_key_value_args(['embed={"$type":"blob"}'])
        {'embed': {'$type': 'blob'}}
    """
    result: dict[str, Any] = {}
    for arg in args:
        if "=" not in arg:
            console.print(f"[red]error:[/red] invalid argument format: {arg}")
            console.print("[dim]use key=value format[/dim]")
            sys.exit(1)
        key, value = arg.split("=", 1)

        # try JSON first for objects/arrays
        if value.startswith(("{", "[")):
            try:
                result[key] = json.loads(value)
                continue
            except json.JSONDecodeError as e:
                console.print(f"[red]error:[/red] invalid JSON for {key}: {e}")
                sys.exit(1)

        # parse primitives
        if value.lower() == "true":
            result[key] = True
        elif value.lower() == "false":
            result[key] = False
        elif value.lower() == "null":
            result[key] = None
        elif value.isdigit():
            result[key] = int(value)
        else:
            try:
                result[key] = float(value)
            except ValueError:
                result[key] = value
    return result



"""tests for argument parsing."""

from __future__ import annotations

import pytest

from pdsx._internal.parsing import parse_key_value_args


def test_parse_string_values() -> None:
    """test parsing string values."""
    result = parse_key_value_args(["name=test", "description=hello world"])
    assert result == {"name": "test", "description": "hello world"}


def test_parse_boolean_values() -> None:
    """test parsing boolean values."""
    result = parse_key_value_args(["active=true", "disabled=false"])
    assert result == {"active": True, "disabled": False}


def test_parse_null_values() -> None:
    """test parsing null values."""
    result = parse_key_value_args(["value=null"])
    assert result == {"value": None}


def test_parse_integer_values() -> None:
    """test parsing integer values."""
    result = parse_key_value_args(["count=42", "limit=100"])
    assert result == {"count": 42, "limit": 100}


def test_parse_float_values() -> None:
    """test parsing float values."""
    result = parse_key_value_args(["rate=3.14", "score=99.9"])
    assert result == {"rate": 3.14, "score": 99.9}


def test_parse_mixed_values() -> None:
    """test parsing mixed value types."""
    result = parse_key_value_args(
        [
            "name=test",
            "count=5",
            "active=true",
            "value=null",
            "rate=2.5",
        ]
    )
    assert result == {
        "name": "test",
        "count": 5,
        "active": True,
        "value": None,
        "rate": 2.5,
    }


def test_invalid_format_raises_system_exit() -> None:
    """test that invalid format raises SystemExit."""
    with pytest.raises(SystemExit):
        parse_key_value_args(["invalid"])


def test_parse_json_object() -> None:
    """test parsing JSON object values."""
    result = parse_key_value_args(['embed={"$type":"blob","size":123}'])
    assert result == {"embed": {"$type": "blob", "size": 123}}


def test_parse_json_array() -> None:
    """test parsing JSON array values."""
    result = parse_key_value_args(['tags=["python","atproto"]'])
    assert result == {"tags": ["python", "atproto"]}


def test_parse_complex_nested_json() -> None:
    """test parsing complex nested JSON."""
    result = parse_key_value_args(
        [
            'reply={"root":{"uri":"at://...","cid":"..."},"parent":{"uri":"at://...","cid":"..."}}'
        ]
    )
    assert result == {
        "reply": {
            "root": {"uri": "at://...", "cid": "..."},
            "parent": {"uri": "at://...", "cid": "..."},
        }
    }


def test_parse_mixed_primitive_and_json() -> None:
    """test parsing mix of primitives and JSON values."""
    result = parse_key_value_args(
        ["text=hello", "count=5", 'meta={"version":1}', "active=true"]
    )
    assert result == {
        "text": "hello",
        "count": 5,
        "meta": {"version": 1},
        "active": True,
    }


def test_invalid_json_raises_system_exit() -> None:
    """test that invalid JSON raises system exit."""
    with pytest.raises(SystemExit):
        parse_key_value_args(["embed={invalid json}"])
